<?php

namespace Drupal\Tests\config_overlay\Functional;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Config\StorageComparer;
use Drupal\Tests\BrowserTestBase;

/**
 * Provides a base test class for testing Configuration Overlay.
 *
 * @group config_overlay
 */
class ConfigOverlayTestingTest extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['config_overlay'];

  /**
   * The configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configStorage;

  /**
   * The configuration synchronization storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configSyncStorage;

  /**
   * The configuration synchronization directory.
   *
   * @var string
   */
  protected $configSyncDirectory;

  /**
   * The serializer.
   *
   * @var \Drupal\Component\Serialization\SerializationInterface
   */
  protected $serializer;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function setUp() {
    parent::setUp();

    $this->configStorage = $this->container->get('config.storage');
    $this->configSyncStorage = $this->container->get('config.storage.sync');
    $this->configSyncDirectory = config_get_config_directory(CONFIG_SYNC_DIRECTORY);
    $this->serializer = $this->container->get('serialization.yaml');
    $this->moduleHandler = $this->container->get('module_handler');
    $this->configManager = $this->container->get('config.manager');
    $this->entityTypeManager = $this->container->get('entity_type.manager');
  }

  /**
   * Tests configuration export after profile installation.
   *
   * @throws \Exception
   */
  public function testConfigExport() {
    $this->doTestInitialConfig();

    // Recreate the menu with the same values (but for the UUID). This will not
    // be detected as a change.
    $menu_storage = $this->entityTypeManager->getStorage('menu');
    /* @var \Drupal\system\MenuInterface $initial_menu */
    $initial_menu = $menu_storage->load('account');
    $recreated_menu = $this->doTestRecreateInitial($initial_menu);

    $this->assertEquals('User account menu', $recreated_menu->label());
    $this->doTestEdit($recreated_menu);
    // Edit the menu again, to make sure that this is detected as an update.
    $this->doTestEdit($recreated_menu);
    $this->doTestRecreateAgain($recreated_menu);
  }

  /**
   * Checks that the configuration export is correct after installation.
   */
  protected function doTestInitialConfig() {
    $files = $this->exportConfig();
    foreach ($this->getExpectedConfig() as $name => $config) {
      $this->assertArrayHasKey($name, $files);

      $uri = $files[$name]->uri;
      $this->assertSame($config, $this->readConfigFile($uri), 'Configuration name: ' . $name);

      unset($files[$name]);
    }

    // This is functionally equivalent to Inspector::assertAllNotEmpty(), but
    // yields a more expressive error message in case of failure.
    foreach ($files as $name => $file) {
      $this->assertSame([], $this->readConfigFile($file->uri), 'Unexpected configuration: ' . $name);
    }

    // Make sure that the configuration storage comparer detects no changes.
    $this->assertConfigStorageChanges();
  }

  /**
   * Tests recreating a shipped configuration entity.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityInterface $initial_entity
   *   The entity to recreate.
   *
   * @return \Drupal\Core\Config\Entity\ConfigEntityInterface
   *   The recreated entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function doTestRecreateInitial(ConfigEntityInterface $initial_entity) {
    // Change some configuration and make sure that it is detected correctly.
    $config_name = $initial_entity->getConfigDependencyName();
    $files = $this->exportConfig();
    $this->assertArrayNotHasKey($config_name, $files);

    $recreated_menu = $this->recreateEntity($initial_entity);

    $this->assertConfigStorageChanges();
    $files = $this->exportConfig();
    $this->assertArrayNotHasKey($config_name, $files);

    return $recreated_menu;
  }

  /**
   * Tests editing a configuration entity.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityInterface $entity
   *   The entity to edit.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function doTestEdit(ConfigEntityInterface $entity) {
    // Edit the menu, so that it will be exported to the synchronization
    // directory.
    $label_key = $entity->getEntityType()->getKey('label');
    $entity
      ->set($label_key, $entity->label() . ' EDITED')
      ->save();

    $config_name = $entity->getConfigDependencyName();
    $this->assertConfigStorageChanges([], [$config_name]);
    $files = $this->exportConfig();
    $this->assertArrayHasKey($config_name, $files);
    $config = $this->readConfigFile($files[$config_name]->uri);
    $this->assertArrayHasKey($label_key, $config);
    $this->assertEquals($entity->label(), $config[$label_key]);
    $this->assertArrayHasKey('uuid', $config);
  }

  /**
   * Tests recreating a non-shipped configuration entity.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityInterface $recreated_entity
   *   The entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function doTestRecreateAgain(ConfigEntityInterface $recreated_entity) {
    // Recreate the menu again with the same values (but for the UUID). Since
    // the menu has been exported with its UUID, this will now be detected as
    // a recreation.
    $this->recreateEntity($recreated_entity);

    $config_name = $recreated_entity->getConfigDependencyName();
    $this->assertConfigStorageChanges([$config_name], [], [$config_name]);
    $files = $this->exportConfig();
    $this->assertArrayHasKey($config_name, $files);
    $rerecreated_menu_config = $this->readConfigFile($files[$config_name]->uri);
    $this->assertArrayHasKey('uuid', $rerecreated_menu_config);
    $this->assertNotEquals($recreated_entity->uuid(), $rerecreated_menu_config['uuid']);
    $this->assertArrayHasKey('id', $rerecreated_menu_config);
    $this->assertEquals($recreated_entity->id(), $rerecreated_menu_config['id']);
  }

  /**
   * Exports configuration for all collections and returns the exported files.
   *
   * @return \stdClass[]
   *   An associative array of objects with 'uri', 'filename', and 'name'
   *   properties corresponding to the matched files. The array is keyed by the
   *   'name' property of the respective files.
   */
  protected function exportConfig() {
    $comparer = new StorageComparer($this->configStorage, $this->configSyncStorage, $this->configManager);
    $comparer->createChangelist();
    foreach ($comparer->getAllCollectionNames() as $collection) {
      $active_storage = $this->configStorage->createCollection($collection);
      $sync_storage = $this->configSyncStorage->createCollection($collection);

      foreach ($comparer->getChangelist('delete', $collection) as $config_name_to_delete) {
        $sync_storage->delete($config_name_to_delete);
      }

      foreach ($comparer->getChangelist('rename', $collection) as $rename) {
        list($old_name, $new_name) = explode('::', $rename);
        $sync_storage->rename($old_name, $new_name);
      }

      $config_names_to_write = array_merge(
        $comparer->getChangelist('create', $collection),
        $comparer->getChangelist('update', $collection)
      );
      $active_config = $active_storage->readMultiple($config_names_to_write);
      foreach ($config_names_to_write as $config_name_to_write) {
        $sync_storage->write($config_name_to_write, $active_config[$config_name_to_write]);
      }
    }

    $extension = $this->serializer->getFileExtension();
    return file_scan_directory($this->configSyncDirectory, "/.\.$extension$/", ['key' => 'name']);
  }

  /**
   * Recreates a given configuration entity.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityInterface $entity
   *   The entity.
   *
   * @return \Drupal\Core\Config\Entity\ConfigEntityInterface
   *   The recreated entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function recreateEntity(ConfigEntityInterface $entity) {
    $entity->delete();
    $entity_storage = $this->entityTypeManager->getStorage($entity->getEntityTypeId());
    $values = $entity->toArray();
    unset($values['uuid']);
    $recreated_entity = $entity_storage->create($values);
    $recreated_entity->save();
    return $recreated_entity;
  }

  /**
   * Gets an array of expected configuration that will be exported.
   *
   * Due to ExtensionConfigFilter this should only contain files that have
   * overridden configuration. For those it should contain the entire
   * configuration file.
   *
   * @return array[]
   *   An array of expected configuration where the keys are the configuration
   *   names and the values are arrays which contain the expected configuration.
   *
   * @see \Drupal\Tests\config_overlay\Functional\ConfigOverlayTestingTest::getOverriddenConfig()
   * @see \Drupal\Tests\config_overlay\Functional\ConfigOverlayTestingTest::getUri()
   */
  protected function getExpectedConfig() {
    $overridden_config = $this->getOverriddenConfig();

    $initial_config = [];
    // Read the initial config for all configuration that is overridden and
    // merge in the overridden values below.
    foreach (array_keys($overridden_config) as $config_name) {
      if ($uri = $this->getUri($config_name)) {
        $data = $this->readConfigFile($uri);
        $initial_config[$config_name] = $this->processConfigData($config_name, $data);
      }
    }

    return NestedArray::mergeDeepArray([$initial_config, $overridden_config], TRUE);
  }

  /**
   * Returns the overridden configuration for this test.
   *
   * @return array[]
   *   An array of overridden configuration where the keys are the configuration
   *   names and the values are arrays which contain the overridden portions of
   *   the configuration.
   */
  protected function getOverriddenConfig() {
    $site_config = $this->configStorage->read('system.site');

    return [
      'core.extension' => [
        'module' => [
          'config_filter' => 0,
          'config_overlay' => 0,
          'dynamic_page_cache' => 0,
          'page_cache' => 0,
          'system' => 0,
          'user' => 0,
          $this->profile => 1000,
        ],
        'theme' => [
          'stable' => 0,
          'classy' => 0,
        ],
        'profile' => $this->profile,
      ],
      /* @see install_base_system() */
      'system.file' => [
        'path' => [
          'temporary' => $this->tempFilesDirectory,
        ],
      ],
      /* @see \Drupal\Core\Test\FunctionalTestSetupTrait::initConfig() */
      'system.date' => [
        'timezone' => [
          'default' => 'Australia/Sydney',
        ],
      ],
      'system.logging' => [
        'error_level' => ERROR_REPORTING_DISPLAY_VERBOSE,
      ],
      'system.mail' => [
        'interface' => [
          'default' => 'test_mail_collector',
        ],
      ],
      'system.performance' => [
        'css' => [
          'preprocess' => FALSE,
        ],
        'js' => [
          'preprocess' => FALSE,
        ],
      ],
      'system.site' => [
        'uuid' => $site_config['uuid'],
        /* @see \Drupal\Core\Test\FunctionalTestSetupTrait::installParameters() */
        'name' => 'Drupal',
        'mail' => 'simpletest@example.com',
      ],
    ];
  }

  /**
   * Gets the file URI for the given configuration name.
   *
   * @param string $config_name
   *   The configuration name to return the file URI for.
   *
   * @return string|false
   *   The file URI for the given configuration, or FALSE if no configuration
   *   file was found.
   */
  protected function getUri($config_name) {
    $extension = $this->serializer->getFileExtension();

    $base_directory = NULL;
    // Reverse the list of directories so that the profile directory comes first
    // so that any profile-provided configuration will be used instead of
    // the respective module-provided configuration.
    $directories = array_reverse($this->moduleHandler->getModuleDirectories());
    $directories[] = $this->root . '/core';

    foreach ($directories as $directory) {
      foreach (['install', 'optional'] as $subdirectory) {
        $uri = $directory . "/config/$subdirectory/$config_name.$extension";
        if (file_exists($uri)) {
          return $uri;
        }
      }
    }

    return FALSE;
  }

  /**
   * Reads a single configuration file.
   *
   * @param string $uri
   *   The URI of the configuration file.
   *
   * @return array|null
   *   An array of configuration data contained in the file or NULL if the file
   *   is empty.
   */
  protected function readConfigFile($uri) {
    return $this->serializer->decode(file_get_contents($uri));
  }

  /**
   * Processes configuration data so that it matches the exported state.
   *
   * @param string $config_name
   *   The configuration name.
   * @param array $data
   *   The configuration data.
   *
   * @return array
   *   The processed configuration data.
   */
  protected function processConfigData($config_name, array $data) {
    // Add the default config hash.
    /* @see \Drupal\Core\Config\ConfigInstaller::createConfiguration() */
    $data['_core'] = ['default_config_hash' => Crypt::hashBase64(serialize($data))];
    if ($this->configManager->getEntityTypeIdByName($config_name)) {
      $data = $this->processConfigEntityData($config_name, $data);
    }
    return $data;
  }

  /**
   * Processes configuration entity data so that it matches the exported state.
   *
   * @param string $config_name
   *   The configuration name.
   * @param array $data
   *   The configuration entity data.
   *
   * @return array
   *   The processed configuration entity data.
   */
  protected function processConfigEntityData($config_name, array $data) {
    // Add the UUID from the active configuration.
    $top_data = ['uuid' => $this->configStorage->read($config_name)['uuid']];

    // Configuration entities are ordered in a particular way when
    // exported, so we need to recreate that here.
    $top_properties = [
      'langcode',
      'status',
      'dependencies',
      'third_party_settings',
      '_core',
    ];
    foreach ($top_properties as $property) {
      if (isset($data[$property])) {
        $top_data[$property] = $data[$property];
      }
    }

    return $top_data + $data;
  }

  /**
   * Reads an optional configuration file from a module.
   *
   * @param string $module_name
   *   The name of the module with the optional configuration.
   * @param string $config_name
   *   The configuration name.
   *
   * @return array
   *   An array of configuration data.
   */
  protected function readOptionalConfig($module_name, $config_name) {
    $module = $this->moduleHandler->getModule($module_name);
    $module_path = $this->root . '/' . $module->getPath();

    $data = $this->readConfigFile("$module_path/config/optional/$config_name.yml");
    return $this->processConfigData($config_name, $data);
  }

  /**
   * Asserts configuration storage changes.
   *
   * @param array $create
   *   An array of names of created configuration.
   * @param array $update
   *   An array of names of updated configuration.
   * @param array $delete
   *   An array of names of deleted configuration.
   * @param array $rename
   *   An array of renamed configuration in the format "old_name::new_name".
   */
  protected function assertConfigStorageChanges(array $create = [], array $update = [], array $delete = [], array $rename = []) {
    $expected = [
      'create' => $create,
      'update' => $update,
      'delete' => $delete,
      'rename' => $rename,
    ];
    $comparer = new StorageComparer($this->configStorage, $this->configSyncStorage, $this->configManager);
    $this->assertEquals($expected, $comparer->createChangelist()->getChangelist());
  }

}
